@@ -34,6 +34,7 @@ Reads and writes JPEG images
 #include "jerror.h"
 #include <errno.h>
 #include <stdlib.h>
+#include <math.h>
 
 #define JPEG_APP13       0xED    /* APP13 marker code */
 #define JPEG_APP1 (JPEG_APP0 + 1)
@@ -46,6 +47,10 @@ Reads and writes JPEG images
 
 static unsigned char fake_eoi[]={(JOCTET) 0xFF,(JOCTET) JPEG_EOI};
 
+#ifdef JPEG_C_PARAM_SUPPORTED
+#define IS_MOZJPEG
+#endif
+
 /* Source and Destination managers */
 
 typedef struct {
@@ -237,7 +242,7 @@ wiol_term_destination (j_compress_ptr cinfo) {
     ERREXIT(cinfo, JERR_FILE_WRITE);
   }
 
-  if (dest != NULL) myfree(dest->buffer);
+  myfree(dest->buffer);
 }
 
 
@@ -340,10 +345,12 @@ transfer_gray(i_color *out, JSAMPARRAY in, int width) {
 typedef void (*transfer_function_t)(i_color *out, JSAMPARRAY in, int width);
 
 static const char version_string[] =
-#ifdef LIBJPEG_TURBO_VERSION
-  "libjpeg-turbo " STRINGIFY(LIBJPEG_TURBO_VERSION) " api " STRINGIFY(JPEG_LIB_VERSION)
+#if defined(IS_MOZJPEG)
+  "mozjpeg version " STRINGIFY(LIBJPEG_TURBO_VERSION) " api " STRINGIFY(JPEG_LIB_VERSION)
+#elif defined(LIBJPEG_TURBO_VERSION)
+  "libjpeg-turbo version " STRINGIFY(LIBJPEG_TURBO_VERSION) " api " STRINGIFY(JPEG_LIB_VERSION)
 #else
-  "libjpeg " STRINGIFY(JPEG_LIB_VERSION)
+  "libjpeg version " STRINGIFY(JPEG_LIB_VERSION) " api " STRINGIFY(JPEG_LIB_VERSION)
 #endif
   ;
 
@@ -358,6 +365,42 @@ i_libjpeg_version(void) {
   return version_string;
 }
 
+int
+is_turbojpeg(void) {
+#if defined(LIBJPEG_TURBO_VERSION)
+  return 1;
+#else
+  return 0;
+#endif
+}
+
+int
+is_mozjpeg(void) {
+#ifdef IS_MOZJPEG
+  return 1;
+#else
+  return 0;
+#endif
+}
+
+int
+has_encode_arith_coding(void) {
+#ifdef C_ARITH_CODING_SUPPORTED
+  return 1;
+#else
+  return 0;
+#endif
+}
+
+int
+has_decode_arith_coding(void) {
+#ifdef D_ARITH_CODING_SUPPORTED
+  return 1;
+#else
+  return 0;
+#endif
+}
+
 /*
 =item i_readjpeg_wiol(data, length, iptc_itext, itlength)
 
@@ -376,6 +419,7 @@ i_readjpeg_wiol(io_glue *data, int length, char** iptc_itext, int *itlength) {
   transfer_function_t transfer_f;
   int channels;
   volatile int src_set = 0;
+  int jfif;
 
   mm_log((1,"i_readjpeg_wiol(data %p, length %d,iptc_itext %p)\n", data, length, iptc_itext));
 
@@ -511,7 +555,9 @@ i_readjpeg_wiol(io_glue *data, int length, char** iptc_itext, int *itlength) {
 
   i_tags_setn(&im->tags, "jpeg_out_color_space", cinfo.out_color_space);
   i_tags_setn(&im->tags, "jpeg_color_space", cinfo.jpeg_color_space);
+  i_tags_setn(&im->tags, "jpeg_read_arithmetic", cinfo.arith_code);
 
+  i_tags_setn(&im->tags, "jpeg_read_jfif", !!cinfo.saw_JFIF_marker);
   if (cinfo.saw_JFIF_marker) {
     double xres = cinfo.X_density;
     double yres = cinfo.Y_density;
@@ -554,6 +600,42 @@ i_readjpeg_wiol(io_glue *data, int length, char** iptc_itext, int *itlength) {
   return im;
 }
 
+struct moz_option {
+  const char *name;
+#ifdef IS_MOZJPEG
+  unsigned tag;
+#endif
+};
+
+#ifdef IS_MOZJPEG
+#define MOZ_OPTION_ENTRY(name, tag) { name, tag }
+#else
+#define MOZ_OPTION_ENTRY(name, tag) { name }
+#endif
+
+static struct moz_option boolean_options[] = {
+  MOZ_OPTION_ENTRY("jpeg_optimize_scans",        JBOOLEAN_OPTIMIZE_SCANS),
+  MOZ_OPTION_ENTRY("jpeg_trellis_quant",         JBOOLEAN_TRELLIS_QUANT),
+  MOZ_OPTION_ENTRY("jpeg_trellis_quant_dc",      JBOOLEAN_TRELLIS_QUANT_DC),
+  MOZ_OPTION_ENTRY("jpeg_tresllis_eob_opt",      JBOOLEAN_TRELLIS_EOB_OPT),
+  MOZ_OPTION_ENTRY("jpeg_use_lambda_weight_tbl", JBOOLEAN_USE_LAMBDA_WEIGHT_TBL),
+  MOZ_OPTION_ENTRY("jpeg_use_scans_in_trellis",  JBOOLEAN_USE_SCANS_IN_TRELLIS),
+  MOZ_OPTION_ENTRY("jpeg_overshoot_deringing",   JBOOLEAN_OVERSHOOT_DERINGING)
+};
+
+static struct moz_option float_options[] = {
+  MOZ_OPTION_ENTRY("jpeg_lambda_log_scale1",       JFLOAT_LAMBDA_LOG_SCALE1),
+  MOZ_OPTION_ENTRY("jpeg_lambda_log_scale2",       JFLOAT_LAMBDA_LOG_SCALE2),
+  MOZ_OPTION_ENTRY("jpeg_trellis_delta_dc_weight", JFLOAT_TRELLIS_DELTA_DC_WEIGHT)
+};
+
+static struct moz_option int_options[] = {
+  MOZ_OPTION_ENTRY("jpeg_trellis_freq_split", JINT_TRELLIS_FREQ_SPLIT),
+  MOZ_OPTION_ENTRY("jpeg_trellis_num_loops",  JINT_TRELLIS_NUM_LOOPS),
+  MOZ_OPTION_ENTRY("jpeg_base_quant_tbl_idx", JINT_BASE_QUANT_TBL_IDX),
+  MOZ_OPTION_ENTRY("jpeg_dc_scan_opt_mode",   JINT_DC_SCAN_OPT_MODE)
+};
+
 /*
 =item i_writejpeg_wiol(im, ig, qfactor)
 
@@ -563,13 +645,15 @@ i_readjpeg_wiol(io_glue *data, int length, char** iptc_itext, int *itlength) {
 undef_int
 i_writejpeg_wiol(i_img *im, io_glue *ig, int qfactor) {
   JSAMPLE *image_buffer;
-  int quality;
   int got_xres, got_yres, aspect_only, resunit;
   double xres, yres;
   int comment_entry;
   int want_channels = im->channels;
   int progressive = 0;
   int optimize = 0;
+  int arithmetic = 0;
+  char profile_name[20] = "";
+  int jfif;
 
   struct jpeg_compress_struct cinfo;
   struct my_error_mgr jerr;
@@ -591,7 +675,6 @@ i_writejpeg_wiol(i_img *im, io_glue *ig, int qfactor) {
   if (!(im->channels==1 || im->channels==3)) { 
     want_channels = im->channels - 1;
   }
-  quality = qfactor;
 
   cinfo.err = jpeg_std_error(&jerr.pub);
   jerr.pub.error_exit = my_error_exit;
@@ -600,6 +683,7 @@ i_writejpeg_wiol(i_img *im, io_glue *ig, int qfactor) {
   jpeg_create_compress(&cinfo);
 
   if (setjmp(jerr.setjmp_buffer)) {
+  fail:
     jpeg_destroy_compress(&cinfo);
     if (data)
       myfree(data);
@@ -623,8 +707,146 @@ i_writejpeg_wiol(i_img *im, io_glue *ig, int qfactor) {
     cinfo.in_color_space = JCS_GRAYSCALE; 	/* colorspace of input image */
   }
 
+  if (i_tags_get_string(&im->tags, "jpeg_compress_profile", 0,
+                        profile_name, sizeof(profile_name))) {
+    if (strcmp(profile_name, "fastest") == 0) {
+#ifdef IS_MOZJPEG
+      jpeg_c_set_int_param(&cinfo, JINT_COMPRESS_PROFILE, JCP_FASTEST);
+#endif
+      /* else default */
+    }
+    else if (strcmp(profile_name, "max") == 0) {
+#ifdef IS_MOZJPEG
+      jpeg_c_set_int_param(&cinfo, JINT_COMPRESS_PROFILE, JCP_MAX_COMPRESSION);
+#else
+      i_push_error(0, "jpeg_compress_profile=max requires mozjpeg");
+      goto fail;
+#endif
+    }
+    else {
+      i_push_errorf(0, "jpeg_compress_profile=%s unknown", profile_name);
+      goto fail;
+    }
+  }
+#ifdef IS_MOZJPEG
+  else {
+    jpeg_c_set_int_param(&cinfo, JINT_COMPRESS_PROFILE, JCP_FASTEST);
+  }
+#endif
+
   jpeg_set_defaults(&cinfo);
-  jpeg_set_quality(&cinfo, quality, TRUE);  /* limit to baseline-JPEG values */
+
+  {
+    char tune_str[20];
+    if (i_tags_get_string(&im->tags, "jpeg_tune", 0, tune_str, sizeof(tune_str))) {
+      if (strcmp(tune_str, "psnr") == 0) {
+#ifdef IS_MOZJPEG
+        jpeg_c_set_int_param(&cinfo, JINT_BASE_QUANT_TBL_IDX, 1);
+        jpeg_c_set_float_param(&cinfo, JFLOAT_LAMBDA_LOG_SCALE1, 9.0);
+        jpeg_c_set_float_param(&cinfo, JFLOAT_LAMBDA_LOG_SCALE2, 0.0);
+        jpeg_c_set_bool_param(&cinfo, JBOOLEAN_USE_LAMBDA_WEIGHT_TBL, FALSE);
+#endif
+      }
+      else if (strcmp(tune_str, "ssim") == 0) {
+#ifdef IS_MOZJPEG
+        jpeg_c_set_int_param(&cinfo, JINT_BASE_QUANT_TBL_IDX, 1);
+        jpeg_c_set_float_param(&cinfo, JFLOAT_LAMBDA_LOG_SCALE1, 11.5);
+        jpeg_c_set_float_param(&cinfo, JFLOAT_LAMBDA_LOG_SCALE2, 12.75);
+        jpeg_c_set_bool_param(&cinfo, JBOOLEAN_USE_LAMBDA_WEIGHT_TBL, FALSE);
+#endif
+      }
+      else if (strcmp(tune_str, "ms-ssim") == 0) {
+#ifdef IS_MOZJPEG
+        jpeg_c_set_int_param(&cinfo, JINT_BASE_QUANT_TBL_IDX, 3);
+        jpeg_c_set_float_param(&cinfo, JFLOAT_LAMBDA_LOG_SCALE1, 12.0);
+        jpeg_c_set_float_param(&cinfo, JFLOAT_LAMBDA_LOG_SCALE2, 13.0);
+        jpeg_c_set_bool_param(&cinfo, JBOOLEAN_USE_LAMBDA_WEIGHT_TBL, TRUE);
+#endif
+      }
+      else if (strcmp(tune_str, "hvs-psnr") == 0) {
+#ifdef IS_MOZJPEG
+        jpeg_c_set_int_param(&cinfo, JINT_BASE_QUANT_TBL_IDX, 3);
+        jpeg_c_set_float_param(&cinfo, JFLOAT_LAMBDA_LOG_SCALE1, 14.75);
+        jpeg_c_set_float_param(&cinfo, JFLOAT_LAMBDA_LOG_SCALE2, 16.5);
+        jpeg_c_set_bool_param(&cinfo, JBOOLEAN_USE_LAMBDA_WEIGHT_TBL, TRUE);
+#endif
+      }
+      else {
+        i_push_errorf(0, "unknown value '%s' for jpeg_tune", tune_str);
+        goto fail;
+      }
+#ifndef IS_MOZJPEG
+      i_push_error(0, "jpeg_tune requires Imager::File::JPEG be built with mozjpeg");
+      goto fail;
+#endif
+    }
+  }
+
+  jpeg_set_quality(&cinfo, qfactor, TRUE);  /* limit to baseline-JPEG values */
+
+  if (i_tags_get_int(&im->tags, "jpeg_jfif", 0, &jfif) && !jfif) {
+    cinfo.write_JFIF_header = 0;
+  }
+
+  {
+    int i;
+    for (i = 0; i < sizeof(boolean_options) / sizeof(boolean_options[0]); ++i) {
+      int val;
+      const struct moz_option *opt = boolean_options + i;
+      if (i_tags_get_int(&im->tags, opt->name, 0, &val)) {
+#ifdef IS_MOZJPEG
+	jpeg_c_set_bool_param(&cinfo, opt->tag, !!val);
+#else
+	i_push_errorf(0, "option %s requires Imager::File::JPEG be built with mozjpeg",
+		      opt->name);
+	goto fail;
+#endif
+      }
+    }
+  }
+
+  {
+    int i;
+    for (i = 0; i < sizeof(float_options) / sizeof(float_options[0]); ++i) {
+      double val;
+      const struct moz_option *opt = float_options + i;
+      if (i_tags_get_float(&im->tags, opt->name, 0, &val)) {
+#ifdef IS_MOZJPEG
+	float f = val;
+	float g;
+	jpeg_c_set_float_param(&cinfo, opt->tag, f);
+	g = jpeg_c_get_float_param(&cinfo, opt->tag);
+	if (fabs(g-f) > 0.0001) {
+	  i_push_errorf(0, "invalid value %g for tag %s", val, opt->name);
+	}
+#else
+	i_push_errorf(0, "option %s requires Imager::File::JPEG be built with mozjpeg",
+		      opt->name);
+	goto fail;
+#endif
+      }
+    }
+  }
+
+    {
+    int i;
+    for (i = 0; i < sizeof(int_options) / sizeof(int_options[0]); ++i) {
+      int val;
+      const struct moz_option *opt = int_options + i;
+      if (i_tags_get_int(&im->tags, opt->name, 0, &val)) {
+#ifdef IS_MOZJPEG
+	jpeg_c_set_int_param(&cinfo, opt->tag, val);
+	if (jpeg_c_get_int_param(&cinfo, opt->tag) != val) {
+	  i_push_errorf(0, "invalid value %d for tag %s", val, opt->name);
+	}
+#else
+	i_push_errorf(0, "option %s requires Imager::File::JPEG be built with mozjpeg",
+		      opt->name);
+	goto fail;
+#endif
+      }
+    }
+  }
 
   if (!i_tags_get_int(&im->tags, "jpeg_progressive", 0, &progressive))
     progressive = 0;
@@ -634,6 +856,9 @@ i_writejpeg_wiol(i_img *im, io_glue *ig, int qfactor) {
   if (!i_tags_get_int(&im->tags, "jpeg_optimize", 0, &optimize))
     optimize = 0;
   cinfo.optimize_coding = optimize;
+  if (i_tags_get_int(&im->tags, "jpeg_arithmetic", 0, &arithmetic)) {
+    cinfo.arith_code = arithmetic != 0;
+  }
 
   got_xres = i_tags_get_float(&im->tags, "i_xres", 0, &xres);
   got_yres = i_tags_get_float(&im->tags, "i_yres", 0, &yres);
@@ -660,6 +885,79 @@ i_writejpeg_wiol(i_img *im, io_glue *ig, int qfactor) {
     cinfo.Y_density = (int)(yres + 0.5);
   }
 
+  {
+    int smooth;
+    if (i_tags_get_int(&im->tags, "jpeg_smooth", 0, &smooth)) {
+      if (smooth < 0 || smooth > 100) {
+        i_push_error(0, "jpeg_smooth must be an integer from 0 to 100");
+        goto fail;
+      }
+      cinfo.smoothing_factor = smooth;
+    }
+  }
+
+  {
+    char restart_str[20];
+    if (i_tags_get_string(&im->tags, "jpeg_restart", 0, restart_str, sizeof(restart_str))) {
+      long restart_count;
+      char block_flag = '\0';
+      if (sscanf(restart_str, "%ld%c", &restart_count, &block_flag)
+          && restart_count >= 0 && restart_count <= 65535
+          && (block_flag == '\0' || block_flag == 'b' || block_flag == 'B')) {
+        if (block_flag) {
+          cinfo.restart_interval = (unsigned)restart_count;
+        }
+        else {
+          cinfo.restart_in_rows = (int)restart_count;
+        }
+      }
+      else {
+        i_push_error(0, "jpeg_restart must be an integer from 0 to 65535 followed by an optional b");
+        goto fail;
+      }
+    }
+  }
+
+  {
+    char sample_str[80];
+
+    if (i_tags_get_string(&im->tags, "jpeg_sample", 0, sample_str, sizeof(sample_str))) {
+      int x, y, n;
+      char sep;
+      char *p = sample_str;
+      int index = 0;
+      while (*p) {
+        if (sscanf(p, "%d%c%d%n", &x, &sep, &y, &n) == 3 &&
+            x >= 1 && x <= 4 && y >= 1 && y <= 4 &&
+            (sep == 'x' || sep == 'X') &&
+            index < MAX_COMPONENTS) {
+          cinfo.comp_info[index].h_samp_factor = x;
+          cinfo.comp_info[index].v_samp_factor = y;
+          ++index;
+        }
+        else {
+        failsample:
+          i_push_error(0, "jpeg_sample: must match /^[1-4]x[1-4](,[1-4]x[1-4]){0,9}$/aai");
+          goto fail;
+        }
+        p += n;
+        if (p[0] == ',') {
+          if (!p[1])
+            goto failsample;
+          ++p;
+        }
+        else if (p[0]) {
+          goto failsample;
+        }
+      }
+      /* fill the rest with 1x1 like cjpeg does */
+      for ( ; index < MAX_COMPONENTS; ++index) {
+        cinfo.comp_info[index].h_samp_factor = x;
+        cinfo.comp_info[index].v_samp_factor = y;
+      }
+    }
+  }
+
   jpeg_start_compress(&cinfo, TRUE);
 
   if (i_tags_find(&im->tags, "jpeg_comment", 0, &comment_entry)) {
@@ -670,7 +968,7 @@ i_writejpeg_wiol(i_img *im, io_glue *ig, int qfactor) {
 
   row_stride = im->xsize * im->channels;	/* JSAMPLEs per row in image_buffer */
 
-  if (!im->virtual && im->type == i_direct_type && im->bits == i_8_bits
+  if (!i_img_virtual(im) && im->type == i_direct_type && im->bits == i_8_bits
       && im->channels == want_channels) {
     image_buffer=im->idata;
 
